﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Common;
using System.Transactions;
using System.Data.SqlClient;

namespace HeuristicLab.Services.Deployment.DataAccess {
  public class PluginStore {

    private PluginStoreClassesDataContext ctx;
    public PluginStore() {
      ctx = new PluginStoreClassesDataContext();
    }

    public IEnumerable<ProductDescription> Products {
      get {
        return from p in ctx.Products
               let plugins = from pair in ctx.ProductPlugins
                             from plugin in ctx.Plugins
                             where pair.ProductId == p.Id
                             where plugin.Id == pair.PluginId
                             select plugin
               select MakeProductDescription(p, plugins);
      }
    }

    private ProductDescription MakeProductDescription(Product p, IQueryable<Plugin> plugins) {
      var desc = new ProductDescription(p.Id, p.Name, new Version(p.Version), from plugin in plugins
                                                                              select MakePluginDescription(plugin));
      return desc;
    }

    private PluginDescription MakePluginDescription(Plugin plugin) {
      var desc = new PluginDescription(plugin.Id, plugin.Name, new Version(plugin.Version), from dep in GetDependencies(plugin)
                                                                                            select MakePluginDescription(dep));
      return desc;
    }

    private IEnumerable<Plugin> GetDependencies(Plugin plugin) {
      return from pair in ctx.Dependencies
             from dependency in ctx.Plugins
             where pair.PluginId == plugin.Id
             where pair.DependencyId == dependency.Id
             select dependency;
    }

    public IEnumerable<PluginDescription> Plugins {
      get {
        return from plugin in ctx.Plugins
               select MakePluginDescription(plugin);
      }
    }

    public byte[] PluginFile(PluginDescription pluginDescription) {
      return GetExistingPlugin(pluginDescription.Name, pluginDescription.Version).PluginPackage.Data.ToArray();
    }

    public void Persist(PluginDescription pluginDescription, byte[] pluginPackage) {
      try {
        using (var transaction = new TransactionScope()) {
          Plugin pluginEntity = InsertOrUpdatePlugin(pluginDescription);
          if (pluginEntity.PluginPackage == null) {
            // insert
            pluginEntity.PluginPackage = MakePluginPackage(pluginDescription, pluginPackage);
          } else {
            // update
            pluginEntity.PluginPackage.Data = pluginPackage;
          }
          ctx.SubmitChanges();
          transaction.Complete();
        }
      }
      catch (SqlException ex) {
        throw new ArgumentException("Something went wrong while trying to persist plugin", ex);
      }
      catch (InvalidOperationException ex) {
        throw new ArgumentException("Something went wrong while trying to persist plugin", ex);
      }
    }

    public void Persist(ProductDescription product) {
      try {
        using (var transaction = new TransactionScope()) {
          InsertOrUpdateProduct(product);
          foreach (var plugin in product.Plugins) {
            InsertOrUpdatePlugin(plugin);
          }
          ctx.SubmitChanges();
          transaction.Complete();
        }
      }
      catch (SqlException ex) {
        throw new ArgumentException("Something went wrong while trying to persist product", ex);
      }
      catch (InvalidOperationException ex) {
        throw new ArgumentException("Something went wrong while trying to persist product", ex);
      }
    }

    private void InsertOrUpdateProduct(ProductDescription product) {
      var productEntity = (from p in ctx.Products
                           where p.Name == product.Name
                           where p.Version == product.Version.ToString()
                           select p).FirstOrDefault() ?? MakeProductFromDescription(product);

      if (productEntity.Id <= 0) {
        ctx.Products.InsertOnSubmit(productEntity);
        ctx.SubmitChanges();
      }

      product.Id = productEntity.Id;

      DeleteOldPlugins(productEntity);

      foreach (var plugin in product.Plugins) {
        var existingPlugin = GetExistingPlugin(plugin.Name, plugin.Version);
        ProductPlugin prodPlugin = new ProductPlugin();
        prodPlugin.PluginId = existingPlugin.Id;
        prodPlugin.ProductId = product.Id;
        ctx.ProductPlugins.InsertOnSubmit(prodPlugin);
      }
    }

    private void DeleteOldPlugins(Product productEntity) {
      var oldPlugins = (from p in ctx.ProductPlugins
                        where p.ProductId == productEntity.Id
                        select p).ToList();
      ctx.ProductPlugins.DeleteAllOnSubmit(oldPlugins);
      ctx.SubmitChanges();
    }

    private Plugin InsertOrUpdatePlugin(PluginDescription pluginDescription) {
      var pluginEntity = (from p in ctx.Plugins
                          where p.Name == pluginDescription.Name
                          where p.Version == pluginDescription.Version.ToString()
                          select p).FirstOrDefault() ?? MakePluginFromDescription(pluginDescription);

      if (pluginEntity.Id <= 0) {
        ctx.Plugins.InsertOnSubmit(pluginEntity);
        ctx.SubmitChanges();
      }
      pluginDescription.Id = pluginEntity.Id;

      DeleteOldDependencies(pluginEntity);

      foreach (var dependency in pluginDescription.Dependencies) {
        var dependencyEntity = GetExistingPlugin(dependency.Name, dependency.Version);
        Dependency d = new Dependency();
        d.PluginId = pluginDescription.Id;
        d.DependencyId = dependencyEntity.Id;
        ctx.Dependencies.InsertOnSubmit(d);
      }
      return pluginEntity;

    }

    private void DeleteOldDependencies(Plugin pluginEntity) {
      var oldDependencies = (from dep in ctx.Dependencies
                             where dep.PluginId == pluginEntity.Id
                             select dep).ToList();

      ctx.Dependencies.DeleteAllOnSubmit(oldDependencies);
      ctx.SubmitChanges();
    }

    private Plugin GetExistingPlugin(string name, Version version) {
      return (from p in ctx.Plugins
              where p.Name == name
              where p.Version == version.ToString()
              select p).Single();
    }

    private Plugin MakePluginFromDescription(PluginDescription pluginDescription) {
      var plugin = new Plugin();
      plugin.Id = pluginDescription.Id;
      plugin.Name = pluginDescription.Name;
      plugin.Version = pluginDescription.Version.ToString();
      return plugin;
    }

    private Product MakeProductFromDescription(ProductDescription desc) {
      var product = new Product();
      product.Id = desc.Id;
      product.Name = desc.Name;
      product.Version = desc.Version.ToString();
      return product;
    }

    private PluginPackage MakePluginPackage(PluginDescription pluginDescription, byte[] pluginPackage) {
      var package = new PluginPackage();
      package.Data = pluginPackage;
      package.PluginId = pluginDescription.Id;
      package.FileName = string.Empty;
      return package;
    }
  }
}
